/*
	SObjectizer 5.
*/

/*!
	\file
	\brief A wrapper for all major SO Run-Time includes.
*/

#if !defined( SO_5_RT_HPP )

#include <so_5/message.hpp>
#include <so_5/agent.hpp>
#include <so_5/coop.hpp>
#include <so_5/state.hpp>
#include <so_5/mhood.hpp>
#include <so_5/mbox.hpp>
#include <so_5/environment.hpp>
#include <so_5/agent_coop_notifications.hpp>
#include <so_5/send_functions.hpp>

#include <so_5/mchain_select.hpp>

#include <so_5/stats/std_names.hpp>
#include <so_5/stats/messages.hpp>

#endif
