/*
 * SObjectizer-5
 */

/*!
 * \file
 * \brief A simple metafunction that always "returns" false.
 *
 * \since
 * v.5.6.0
 */

#if !defined( SO_5_DETAILS_ALWAYS_FALSE_HPP )
#define SO_5_DETAILS_ALWAYS_FALSE_HPP

namespace so_5 {

namespace details {

/*!
 * \brief A simple metafunction that always "returns" false.
 *
 * Can be used to trigger static_assert.
 *
 * \since
 * v.5.6.0
 */
template<typename T>
struct always_false
{
	static constexpr const bool value = false;
};

} /* namespace details */

} /* namespace so_5 */

#endif

