/*
 * SObjectizer-5
 */

/*!
 * \file
 * \since
 * v.5.5.22
 *
 * \brief Definition of optional type.
 */

#if !defined( SO_5_OPTIONAL_HPP )
#define SO_5_OPTIONAL_HPP

#include <optional>

namespace so_5 {

using std::optional;

} /* namespace so_5 */

#endif

